@@ -18,30 +18,22 @@ use File::Basename qw(basename dirname fileparse);
 use File::Copy qw(cp);
 use File::Path qw(mkpath);
 use File::Rsync;
+use File::Rsync::Mirror::Recentfile::Done (); # at least needed by thaw()
 use File::Rsync::Mirror::Recentfile::FakeBigFloat qw(:all);
 use File::Temp;
 use List::Pairwise qw(mapp grepp);
 use List::Util qw(first max);
-use Scalar::Util qw(reftype);
+use Scalar::Util qw(blessed reftype);
 use Storable;
 use Time::HiRes qw();
 use YAML::Syck;
 
-use version; our $VERSION = qv('0.0.5');
+use version; our $VERSION = qv('0.1.0');
 
 =head1 SYNOPSIS
 
-B<!!!! PRE-ALPHA ALERT !!!!>
-
-Nothing in here is believed to be stable, nothing yet intended for
-public consumption. The plan is to provide scripts that act as
-frontends for all the backend functionality. Option and method names
-may still change.
-
-For the rationale see the section BACKGROUND.
-
 The documentation in here is normally not needed because the code is
-meant to be run from several standalone programs. For a quick
+considered to be run from several standalone programs. For a quick
 overview, see the file README.mirrorcpan and the bin/ directory of the
 distribution. For the architectural ideas see the section THE
 ARCHITECTURE OF A COLLECTION OF RECENTFILES below.
@@ -73,6 +65,83 @@ sub new {
     return $self;
 }
 
+=head2 my $obj = CLASS->thaw($statusfile)
+
+Constructor from a statusfile left over from a previous
+rmirror run. See also C<runstatusfile>.
+
+=cut
+
+sub _thaw_without_pathdb {
+    my($self,$file) = @_;
+    open my $fh, $file or die "Can't open '$file': $!";
+    local $/ = "\n";
+    my $in_pathdb = 0;
+    my $tfile = File::Temp->new
+        (
+         TEMPLATE => "Recent-thaw-XXXX",
+         TMPDIR => 1,
+         UNLINK => 0,
+         CLEANUP => 0,
+         SUFFIX => '.dat',
+        );
+    my $template_for_eop;
+    while (<$fh>) {
+        if ($in_pathdb) {
+            if (/$template_for_eop/) {
+                $in_pathdb = 0;
+            }
+        } elsif (/(\s+)-\s*__pathdb\s*:/) {
+            $in_pathdb = 1;
+            my $next_attr = sprintf "^%s\\S", " ?" x length($1);
+            $template_for_eop = qr{$next_attr};
+        }
+        print $tfile $_ unless $in_pathdb;
+    }
+    close $tfile or die "Could not close: $!";
+    my $return = $self->thaw($tfile->filename);
+    $return->_havelostpathdb(1);
+    unlink $tfile->filename;
+    return $return;
+}
+sub thaw {
+    my($self, $file) = @_;
+    die "thaw called without statusfile argument" unless defined $file;
+    unless (-e $file){
+        require Carp;
+        Carp::confess("Alert: statusfile '$file' not found");
+    }
+    require YAML::Syck;
+    my $start = time;
+    my $sleeptime = 0.02;
+    while (not mkdir "$file.lock") {
+        my $err = $!;
+        Time::HiRes::sleep $sleeptime;
+        my $waiting = time - $start;
+        if ($waiting >= 3){
+            warn "*** waiting ($waiting) for lock ($err) ***";
+            $sleeptime = 1;
+        }
+    }
+    my $size = -s $file;
+    my $serialized = YAML::Syck::LoadFile($file);
+    rmdir "$file.lock" or die "Could not rmdir lockfile: $!";
+    my $charged_self = $serialized->{reduced_self};
+    my $class = blessed $self;
+    bless $charged_self, $class;
+    my $rfs = $serialized->{reduced_rfs};
+    my $rfclass = $class . "file"; # "Recent" . "file"
+    my $pathdb = $charged_self->_pathdb;
+    for my $rf (@$rfs) {
+        bless $rf, $rfclass;
+        $rf->_pathdb($pathdb);
+    }
+    $charged_self->_recentfiles($rfs);
+    $charged_self->_principal_recentfile($rfs->[0]);
+    # die "FIXME: thaw all recentfiles from reduced_rfs into _recentfiles as well, watch out for pathdb and rsync";
+    return $charged_self;
+}
+
 =head1 ACCESSORS
 
 =cut
@@ -84,6 +153,8 @@ BEGIN {
         (
          "__pathdb",
          "_dirtymark",            # keeps track of the dirtymark of the recentfiles
+         "_havelostpathdb",       # boolean
+         "_have_written_statusfile", # boolean
          "_logfilefordone",       # turns on _logfile on all DONE
                                   # systems (disk intensive)
          "_max_one_state",        # when we have no time left but want
@@ -92,7 +163,7 @@ BEGIN {
          "_principal_recentfile",
          "_recentfiles",
          "_rsync",
-         "_runstatusfile",        # frequenty dumps all rfs
+         "_runstatusfile",        # occasionally dumps all rfs
          "_verbose",              # internal variable for verbose setter/getter
          "_verboselog",           # internal variable for verboselog setter/getter
         );
@@ -172,9 +243,16 @@ Test this with:
        -remote pause.perl.org::authors/RECENT.recent
        -verbose
 
-Note: all parameters that can be passed to recent_events can also be specified here.
+All parameters that can be passed to
+File:Rsync:Mirror:Recentfile::recent_events() can also be specified
+here.
 
-Note: all data are kept in memory
+One additional option is supported. If C<$Options{callback}> is
+specified, it must be a subref. This sub is called whenever one chunk
+of events is found. The first argument to the callback is a reference
+to the currently accumulated array of events.
+
+Note: all data are kept in memory.
 
 =cut
 
@@ -224,6 +302,9 @@ sub news {
         }
         $before = $res->[-1]{epoch};
         $before = $opt{before} if $opt{before} && _bigfloatlt($opt{before},$before);
+        if (my $sub = $opt{callback}) {
+            $sub->($ret);
+        }
     }
     $ret;
 }
@@ -253,7 +334,7 @@ I<Cnt> is the number of entries in this recentfile.
 
 I<Max> is the highest(first) epoch in this recentfile, rounded.
 
-I<Min> is the lowest(last) epoch in thie recentfile, rounded.
+I<Min> is the lowest(last) epoch in this recentfile, rounded.
 
 I<Span> is the timespan currently covered, rounded.
 
@@ -282,6 +363,8 @@ sub overview {
                  scalar @$re,
                  "Dirtymark",
                  $rf->dirtymark ? sprintf("%.2f",$rf->dirtymark) : "-",
+                 "Produced",
+                 sprintf ("%.2f", $rf->{ORIG}{Producers}{time}||0),
                  "Merged",
                  ($rf->interval eq "Z"
                   ?
@@ -547,10 +630,7 @@ sub rmirror {
 
     my $rfs = $self->recentfiles;
 
-    my $_every_20_seconds = sub {
-        $self->principal_recentfile->seed;
-    };
-    $_every_20_seconds->();
+    $self->principal_recentfile->seed;
     my $_sigint = sub {
         # XXX exit gracefully (reminder)
     };
@@ -581,45 +661,74 @@ sub rmirror {
             $self->_dirtymark($dirtymark);
         }
     }
+    my $rstfile = $self->runstatusfile;
+    unless ($self->_have_written_statusfile) {
+        $self->_rmirror_runstatusfile_write ($rstfile, \%options);
+        $self->_have_written_statusfile(1);
+    }
+    $self->_rmirror_loop($minimum_time_per_loop,\%options);
+}
+
+sub _rmirror_loop {
+    my($self,$minimum_time_per_loop,$options) = @_;
   LOOP: while () {
         my $ttleave = time + $minimum_time_per_loop;
-      RECENTFILE: for my $i (0..$#$rfs) {
-            my $rf = $rfs->[$i];
-            if (my $file = $self->_runstatusfile) {
-                $self->_rmirror_runstatusfile ($file, $i, \%options);
-            }
-            if (time > $ttleave){
-                # Must make sure that one file can get fetched in any case
-                $self->_max_one_state(1);
-            }
-            if ($rf->seeded) {
-                $self->_rmirror_mirror ($i, \%options);
-            } elsif ($rf->uptodate){
-                if ($i < $#$rfs){
-                    $rfs->[$i+1]->done->merge($rf->done);
+        my $rstfile = $self->runstatusfile;
+        my $otherproc = $self->_thaw_without_pathdb ($rstfile);
+        my $pid = fork;
+        if (! defined $pid) {
+            warn "Contention: $!";
+            sleep 0.25;
+            next LOOP;
+        } elsif ($pid) {
+            waitpid($pid,0);
+        } else {
+            $self = $self->thaw ($rstfile);
+            my $rfs = $self->recentfiles;
+            $self->principal_recentfile->seed;
+        RECENTFILE: for my $i (0..$#$rfs) {
+                my $rf = $rfs->[$i];
+                if (time > $ttleave) {
+                    # Must make sure that one file can get fetched in any case
+                    $self->_max_one_state(1);
                 }
-                # no further seed necessary because "every_20_seconds" does it
-                next RECENTFILE;
-            }
-        WORKUNIT: while (time < $ttleave) {
-                if ($rf->uptodate) {
-                    $self->_rmirror_sleep_per_connection ($i);
+                if ($rf->seeded) {
+                    $self->_rmirror_mirror ($i, $options);
+                } elsif ($rf->uptodate) {
+                    if ($i < $#$rfs) {
+                        $rfs->[$i+1]->done->merge($rf->done);
+                    }
+                    # no further seed necessary because "periodic" does it
                     next RECENTFILE;
-                } else {
-                    $self->_rmirror_mirror ($i, \%options);
+                }
+            WORKUNIT: while (time < $ttleave) {
+                    if ($rf->uptodate) {
+                        $self->_rmirror_sleep_per_connection ($i);
+                        next RECENTFILE;
+                    } else {
+                        $self->_rmirror_mirror ($i, $options);
+                    }
+                }
+                if ($self->_max_one_state) {
+                    last RECENTFILE;
                 }
             }
-            if ($self->_max_one_state) {
-                last RECENTFILE;
+            $self->_max_one_state(0);
+            my $exit = 0;
+            if ($rfs->[-1]->uptodate) {
+                $self->_rmirror_cleanup;
             }
+            unless ($options->{loop}) {
+                $exit = 1;
+            }
+            $self->_rmirror_runstatusfile_write ($rstfile, $options);
+            exit if $exit;
+            last LOOP;
         }
-        $self->_max_one_state(0);
-        if ($rfs->[-1]->uptodate) {
-            $self->_rmirror_cleanup;
-            if ($options{loop}) {
-            } else {
-                last LOOP;
-            }
+
+        $otherproc = $self->_thaw_without_pathdb ($rstfile);
+        if (!$options->{loop} && $otherproc && $otherproc->recentfiles->[-1]->uptodate) {
+            last LOOP;
         }
         my $sleep = $ttleave - time;
         if ($sleep > 0.01) {
@@ -627,7 +736,6 @@ sub rmirror {
         } else {
             # negative time not invented yet:)
         }
-        $_every_20_seconds->();
     }
 }
 
@@ -675,27 +783,93 @@ sub _rmirror_cleanup {
         my $thismerged = $rfs->[$i]->merged;
         my $next = $rfs->[$i+1];
         my $nextminmax = $next->minmax;
-        # warn "DEBUG: i[$i] nextminmaxmax[$nextminmax->{max}] thismergedepoch[$thismerged->{epoch}]";
         if (not defined $thismerged->{epoch} or _bigfloatlt($nextminmax->{max},$thismerged->{epoch})){
             $next->seed;
-            # warn sprintf "DEBUG: next iv %s seeded since next-minmax-max[$nextminmax->{max}]lt this-merged-epoch[$thismerged->{epoch}]\n", $next->interval;
         }
     }
 }
 
-sub _rmirror_runstatusfile {
-    my($self, $file, $i, $options) = @_;
-    my $rfs = $self->recentfiles;
+=head2 $file = $obj->runstatusfile ($set)
+
+Getter/setter for C<_runstatusfile> attribute. Defaults to a temporary
+file created by C<File::Temp>. A status file is required for
+C<rmirror> working. Since it may be interesting for debugging
+purposes, you may want to specify a permanent file for this.
+
+=cut
+sub runstatusfile {
+    my($self,$set) = @_;
+    if (defined $set) {
+        $self->_runstatusfile ($set);
+    }
+    my $x = $self->_runstatusfile;
+    unless (defined $x) {
+        require File::Temp;
+        my $tfile = File::Temp->new
+            (
+             TEMPLATE => "Recent-XXXX",
+             TMPDIR => 1,
+             UNLINK => 0,
+             CLEANUP => 0,
+             SUFFIX => '.dat',
+            );
+        $self->_runstatusfile($tfile->filename);
+    }
+    return $self->_runstatusfile;
+}
+
+# unused code.... it was an oops, discovered the thaw() method too
+# late, and starting writing this here....
+sub _rmirror_runstatusfile_read {
+    my($self, $file) = @_;
+
     require YAML::Syck;
+    my $start = time;
+    # XXX is locking useful here?
+    while (not mkdir "$file.lock") {
+        Time::HiRes::sleep 0.2;
+        warn "*** waiting for lock ***" if time - $start >= 3;
+    }
+    my $yml = YAML::Syck::LoadFile $file;
+    rmdir "$file.lock" or die "Could not rmdir lockfile: $!";
+    my $rself = $yml->{reduced_self};
+    my $rfs = $yml->{reduced_rfs};
+    # XXX bring them into self
+}
+
+sub _rmirror_runstatusfile_write {
+    my($self, $file, $options) = @_;
+    my $rself;
+    while (my($k,$v) = each %$self) {
+        next if $k =~ /^-(_principal_recentfile|_recentfiles)$/;
+        $rself->{$k} = $v;
+    }
+    my $rfs = $self->recentfiles;
+    my $rrfs;
+    for my $i (0..$#$rfs) {
+        my $rf = $rfs->[$i];
+        while (my($k,$v) = each %$rf) {
+            next if $k =~ /^-(_current_tempfile_fh|_pathdb|_rsync)$/;
+            $rrfs->[$i]{$k} = $rfs->[$i]{$k};
+        }
+    }
+    require YAML::Syck;
+    my $start = time;
+    while (not mkdir "$file.lock") {
+        Time::HiRes::sleep 0.15;
+        warn "*** waiting for lock ***" if time - $start >= 3;
+    }
     YAML::Syck::DumpFile
           (
-           $file,
-           {i => $i,
+           "$file.new",
+           {
             options => $options,
-            self => [keys %$self], # passing $self leaks, dclone refuses because of globs
             time => time,
-            uptodate => {map {($_=>$rfs->[$_]->uptodate)} 0..$#$rfs},
+            reduced_rfs => $rrfs,
+            reduced_self => $rself,
            });
+    rename "$file.new", $file or die "Could not rename: $!";
+    rmdir "$file.lock" or die "Could not rmdir lockfile: $!";
 }
 
 sub _rmirror_endofloop_sleep {
@@ -774,6 +948,7 @@ sub _fetch_as_tempfile {
         Carp::confess(YAML::Syck::Dump($self->rsync_options));
     }
     my $dst = $fh->filename;
+    local($ENV{LANG}) = "C";
     $rsync->exec
         (
          src => join("/",$self->remoteroot,$rfile),
@@ -923,6 +1098,19 @@ the hosts.
 This is about speeding up rsync operation on large trees. Uses a small
 metadata cocktail and pull technology.
 
+rersyncrecent solves this problem with a couple of (usually 2-10)
+lightweight index files which cover different overlapping time
+intervals. The master writes these files and the clients/slaves can
+construct the full tree from the information contained in them. The
+most recent index file usually covers the last seconds or minutes or
+hours of the tree and depending on the needs, slaves can rsync every
+few seconds or minutes and then bring their trees in full sync.
+
+The rersyncrecent model was developed for CPAN but as it is both
+convenient and economic it is also a general purpose solution. I'm
+looking forward to see a CPAN backbone that is only a few seconds
+behind PAUSE.
+
 =head2 NON-COMPETITORS
 
  File::Mirror        JWU/File-Mirror/File-Mirror-0.10.tar.gz only local trees
@@ -935,6 +1123,7 @@ metadata cocktail and pull technology.
  rsnapshot           www.rsnapshot.org                       focus on backup
  csync               www.csync.org                           more like unison
  multi-rsync         sourceforge 167893                      lan push to many
+ chasm               chasmd.org                              per-directory manifests
 
 =head2 COMPETITORS
 
@@ -963,23 +1152,13 @@ incestuous relation but it has the disadvantage that these batch files
 replicate the contents of the involved files. This seems inappropriate
 when the nodes already have a means of communicating over rsync.
 
+=head2 HONORABLE MENTION
+
 B<instantmirror> at https://fedorahosted.org/InstantMirror/ is an
-ambitious project that tries to combine various technologies to
-overcome the current situation. It's been founded in 2009-03 and at
-the time of this writing it is still a bit early to comment on.
-
-rersyncrecent solves this problem with a couple of (usually 2-10)
-lightweight index files which cover different overlapping time
-intervals. The master writes these files and the clients/slaves can
-construct the full tree from the information contained in them. The
-most recent index file usually covers the last seconds or minutes or
-hours of the tree and depending on the needs, slaves can rsync every
-few seconds or minutes and then bring their trees in full sync.
-
-The rersyncrecent mode was developed for CPAN but as it is convenient
-and economic it is also a general purpose solution. I'm looking
-forward to see a CPAN backbone that is only a few seconds behind
-PAUSE. And then ... the first FUSE based CPAN filesystem anyone?
+ambitious project that tries to combine various technologies (squid,
+bittorrent) to overcome the current slowness with the main focus on
+fedora. It's been founded in 2009-03 and at the time of this writing
+it is still a bit early to comment on.
 
 =head1 LIMITATIONS
 
